Require Import LambdaTamer.LambdaTamer.

Require Import simple_tc_safety.

Set Implicit Arguments.


Module M : SIMPLE_TC_SAFETY.
  Import SimpleTC.

  Inductive typ : Type :=
  | Nil : typ
  | List : typ -> typ
  | ListCons : typ -> typ.

  Definition ty := typ.

  Definition t_nil := Nil.
  Definition t_list := List.
  Definition t_listcons := ListCons.

  Definition env := map ty.
  Definition program_typing := map env.

  Fixpoint all (R : value -> Prop) (v : value) : Prop :=
    match v with
      | value_label l => l = L 0
      | value_cons v1 v2 => R v1 /\ all R v2
    end.

  Fixpoint lr (t : typ) (v : value) {struct t} : Prop :=
    match t with
      | Nil => match v with
                 | value_label l => l = L 0
                 | value_cons _ _ => False
               end
      | List t => all (lr t) v
      | ListCons t => match v with
                        | value_label _ => False
                        | value_cons v1 v2 => lr t v1 /\ all (lr t) v2
                      end
    end.

  Definition subtype (t1 t2 : typ) : Prop :=
    forall v, lr t1 v -> lr t2 v.

  Section subtyping.
    Ltac my_changer :=
      match goal with
        | [ H : context[match ?V with value_label _ => _ | value_cons _ _ => _ end] |- _ ] => destruct V
      end.

    Ltac my_simpler := repeat progress (unfold subtype in *; simpler; try my_changer).
    Ltac my_esimpler := my_simpler; eauto.

    Theorem sub_refl : forall t, subtype t t.
      my_simpler.
    Qed.

    Theorem sub_trans : forall t1 t2 t3,
      subtype t1 t2 ->
      subtype t2 t3 ->
      subtype t1 t3.
      my_simpler.
    Qed.

    Theorem sub_nil  : forall t,
      subtype t_nil (t_list t).
      my_simpler.
    Qed.

    Lemma all_weaken : forall (R1 R2 : value -> Prop) (v : value),
      all R1 v
      -> (forall v, R1 v -> R2 v)
      -> all R2 v.
      induction v; simpler.
    Qed.

    Hint Resolve all_weaken.

    Theorem sub_list : forall t1 t2,
      subtype t1 t2 ->
      subtype (t_list t1) (t_list t2).
      my_esimpler.
    Qed.

    Theorem sub_listmixed : forall t1 t2,
      subtype t1 t2 ->
      subtype (t_listcons t1) (t_list t2).
      my_esimpler.
    Qed.

    Theorem sub_listcons : forall t1 t2,
      subtype t1 t2 ->
      subtype (t_listcons t1) (t_listcons t2).
      my_esimpler.
    Qed.
  End subtyping.

  Hint Resolve sub_refl sub_trans sub_nil
    sub_list sub_listmixed sub_listcons.

  Definition env_sub (e e':env) :=
    forall v t', e'#v = Some t' ->
      exists t, e#v = Some t /\ subtype t t'.

  Definition check_env (s : store) (e : env) : Prop :=
    forall r t, e#r = Some t
      -> exists v, s#r = Some v /\ lr t v.

  Definition typing_dom_match (p: program) (pt: program_typing) : Prop :=
    forall l e, pt#l = Some e -> exists i, p#l = Some i.

  Definition stepstar_plus p s i s' i' :=
    exists n, stepN p (S n) s i s' i'.

  Definition check_instr (p : program) (pt : program_typing) (e : env) (i : instr) (e' : env) : Prop :=
    typing_dom_match p pt
    -> forall s,
      check_env s e
      -> forall i', exists s',
        (stepstar p s (i;;i') s' i'
          /\ check_env s' e')
        \/ exists l, exists i'', exists e'',
          p#l = Some i''
          /\ pt#l = Some e''
          /\ check_env s' e''
          /\ stepstar_plus p s (i;;i') s' i''.

  Hint Constructors step stepstar step_or_halt stepN.

  Inductive Added T (v : T) : Prop := Add : Added v.

  Ltac checkNew E :=
    match goal with
      | [ H : Added ?E' |- _ ] =>
        assert (E = E'); [simpler | idtac]; [idtac]; fail 1
      | _ => generalize (Add E); intro
    end.

  Ltac my_changer :=
    match goal with
      | [ H : context[match ?V with value_label _ => _ | value_cons _ _ => _ end] |- _ ] => destruct V
      | [ |- context[match ?V with value_label _ => _ | value_cons _ _ => _ end] ] => destruct V

      | [ He : check_env ?S ?E, Hr : ?E # ?V = Some _ |- _ ] =>
        checkNew (S # V); generalize (He _ _ Hr)

      | [ Hm : typing_dom_match ?P ?PT, Hl : ?PT # ?L = Some _ |- _ ] =>
        checkNew (P # L); generalize (Hm _ _ Hl)

      | [ Hs : env_sub ?E ?E', Hv : ?E' # ?V = Some _ |- _ ] =>
        checkNew (E # V); generalize (Hs _ _ Hv)

      | [ H1 : ?E = _, H2 : ?E = _ |- _ ] =>
        rewrite H1 in H2
    end.
  
  Ltac my_simpler := repeat progress (unfold subtype, check_instr in *;
    simpler; try my_changer).
  Ltac my_esimpler := my_simpler; eauto.

  Hint Immediate get_set_same.

  Hint Rewrite get_empty get_set_same unset_same : ltamer.
  Hint Rewrite get_set_other unset_other using (eauto 1; fail) : ltamer.

  Lemma stepstar_trans : forall p s1 i1 s2 i2 s3 i3,
    stepstar p s1 i1 s2 i2
    -> stepstar p s2 i2 s3 i3
    -> stepstar p s1 i1 s3 i3.
    induction 1; my_esimpler.
  Qed.

  Lemma stepstar_plus_trans : forall p s1 i1 s2 i2 s3 i3,
    stepstar p s1 i1 s2 i2
    -> stepstar_plus p s2 i2 s3 i3
    -> stepstar_plus p s1 i1 s3 i3.
    unfold stepstar_plus; induction 1; my_esimpler.
  Qed.

  Lemma step_stepstar_plus : forall p s1 i1 s2 i2,
    step p s1 i1 s2 i2
    -> stepstar_plus p s1 i1 s2 i2.
    unfold stepstar_plus; eauto.
  Qed.

  Hint Resolve stepstar_trans stepstar_plus_trans step_stepstar_plus.

  Theorem check_instr_seq : forall p pt e e' e'' i1 i2,
    check_instr p pt e  i1 e' ->
    check_instr p pt e' i2 e'' ->
    check_instr p pt e  (i1;;i2) e''.
    my_simpler.

    match goal with
      | [ Hcheck : forall s, check_env s e -> _,
        He : check_env _ e |- _ ] =>
      generalize (Hcheck _ He (i2;;i')); clear Hcheck
    end; my_simpler; [
      match goal with
        | [ Hcheck : forall s, check_env s ?E -> _,
          He : check_env _ ?E |- _ ] =>
        generalize (Hcheck _ He i'); clear Hcheck
      end; my_simpler; [
        econstructor; left; eauto
        | econstructor; right; eauto 9
      ]
      | econstructor; right; eauto 9
    ].
  Qed.

  Definition value_eq_dec : forall (v1 v2 : value), {v1 = v2} + {v1 <> v2}.
    decide equality; apply label_eq.
  Qed.

  Ltac check_env := red; my_simpler;
    match goal with
      | [ p1 : positive, p2 : positive |- _ ] =>
        destruct (label_eq p1 p2); my_esimpler
    end.

  Lemma env_sub_extend : forall e v t e' s v',
    env_sub (e # v <- t) e'
    -> check_env s e
    -> s # v = Some v'
    -> lr t v'
    -> check_env s e'.
    check_env.
  Qed.

  Lemma env_sub_check : forall e e' s,
    env_sub e e'
    -> check_env s e
    -> check_env s e'.
    red; my_esimpler.
  Qed.

  Lemma env_extend : forall e v t s v',
    check_env s e
    -> lr t v'
    -> check_env (s # v <- v') (e # v <- t).
    check_env.
  Qed.

  Lemma check_unset : forall e v e' s v',
    env_sub (unset ty e v) e'
    -> check_env s e
    -> check_env (s # v <- v') e'.
    check_env.
  Qed.

  Lemma env_sub_refl : forall e,
    env_sub e e.
    red; eauto.
  Qed.

  Hint Resolve env_sub_refl env_sub_check env_sub_extend env_extend check_unset.

  Lemma lr_listcons : forall t x1 x2,
    lr t x1
    -> all (lr t) x2
    -> lr (t_listcons t) (value_cons x1 x2).
    simpler.
  Qed.

  Hint Resolve lr_listcons.

  Theorem check_instr_branch_list : forall p pt e e1 v l t,
    e#v = Some (t_list t) ->
    pt#l = Some e1 ->
    env_sub (e#v <- t_nil) e1 ->
    check_instr p pt
    e (instr_branch_if_nil v l) (e#v <- (t_listcons t)).
    my_simpler;
    match goal with
      | [ s : store, x : value |- _ ] =>
        destruct (value_eq_dec x (value_label (L 0))); my_simpler; exists s; [
          right; eauto 10
          | destruct x; my_simpler; eauto 7
        ]
    end.
  Qed.

  Theorem check_instr_branch_listcons : forall p pt e l v t,
    e#v = Some (t_listcons t) ->
    check_instr p pt
    e (instr_branch_if_nil v l) e.
    my_simpler; eauto 6.
  Qed.

  Theorem check_instr_branch_nil : forall p pt e v l e1 e2,
    e#v = Some t_nil ->
    pt#l = Some e1 ->
    env_sub e e1 ->
    check_instr p pt
    e (instr_branch_if_nil v l) e2.
    my_simpler; exists s; right; eauto 10.
  Qed.

  Lemma step_fetch_field_0' : forall a0 p (r : store) v1 v2 i a1,
    r#v1 = Some (value_cons a0 a1) ->
    step p r (instr_fetch_field v1 0 v2 ;; i) ((r#v2 <- a0) : store) i.
    eauto.
  Qed.

  Lemma step_fetch_field_1' : forall a1 p r v1 v2 i a0,
      r#v1 = Some (value_cons a0 a1) ->
      step p r (instr_fetch_field v1 1 v2 ;; i) (r#v2 <- a1) i.
    eauto.
  Qed.

  Hint Extern 1 (step _ _ _ _ _) =>
    match goal with
      | [ H : context[value_cons ?a0 _] |- _ ] => eapply (step_fetch_field_0' (a0 := a0))
    end.

  Hint Extern 1 (step _ _ _ _ _) =>
    match goal with
      | [ H : context[value_cons _ ?a1] |- _ ] => eapply (step_fetch_field_1' (a1 := a1))
    end.

  Theorem check_instr_fetch_0: forall p pt (e : env) v1 v2 t,
    e#v1 = Some (t_listcons t) ->
    check_instr p pt
    e (instr_fetch_field v1 0 v2) (e#v2 <- t).
    my_simpler; eauto 7.
  Qed.

  Theorem check_instr_fetch_1: forall p pt e v1 v2 t,
    e#v1 = Some (t_listcons t) ->
    check_instr p pt
    e (instr_fetch_field v1 1 v2) (e#v2 <- (t_list t)).
    my_simpler; eauto 7.
  Qed.

  Lemma step_cons' : forall a1 a2 p r v1 v2 v3 i,
    r#v1 = Some a1 ->
    r#v2 = Some a2 ->
    step p r (instr_cons v1 v2 v3 ;; i) (r#v3 <- (value_cons a1 a2)) i.
    eauto.
  Qed.

  Hint Extern 1 (step _ _ _ _ _) =>
    match goal with
      | [ a1 : value, a2 : value |- _ ] => eapply (step_cons' (a1 := a1) (a2 := a2)); eauto; fail
    end.

  Theorem check_instr_cons: forall p pt e v1 v2 v3 t1 t2 t3,
    e#v1 = Some t1 ->
    e#v2 = Some t2 ->
    subtype t1 t3 ->
    subtype t2 (t_list t3) ->
    check_instr p pt
    e (instr_cons v1 v2 v3) (e#v3 <- (t_listcons t3)).
    my_simpler; eauto 9.
  Qed.

  Definition check_block (p : program) (pt : program_typing) (e : env) (i : instr) : Prop :=
    typing_dom_match p pt
    -> forall s,
      check_env s e
      -> exists s', stepstar p s i s' instr_halt
        \/ exists l, exists i', exists e',
          p#l = Some i'
          /\ pt#l = Some e'
          /\ check_env s' e'
          /\ stepstar_plus p s i s' i'.

  Ltac block_simpler := unfold check_block; my_simpler.

  Theorem check_block_halt: forall p pt e,
    check_block p pt e instr_halt.
    block_simpler; eauto.
  Qed.

  Theorem check_block_seq: forall p pt e i1 i2 e',
    check_instr p pt e i1 e' ->
    check_block p pt e' i2 ->
    check_block p pt e (i1 ;; i2).
    block_simpler;
    match goal with
      | [ H1 : forall s : store, _, H2 : check_env _ _ |- _ ] =>
        generalize (H1 _ H2 i2); clear H1; my_simpler
    end; [
      match goal with
        | [ H1 : forall s : store, _, H2 : check_env _ _ |- _ ] =>
          generalize (H1 _ H2); clear H1; my_simpler
      end; [
        eauto
        | econstructor; right; eauto 8
      ]
      | econstructor; right; eauto 8
    ].
  Qed.

  Theorem check_block_jump: forall p pt e v l e1,
    pt#l = Some e1 ->
    env_sub (unset _ e v) e1 ->
    check_block p pt e (instr_getlabel l v ;; instr_jump v).
    block_simpler;
    exists (s # v <- (value_label l)); right; eauto 12.
  Qed.

  Definition check_blocks (p:program) (pt:program_typing) :=
    forall l i,
      p#l = Some i ->
      exists e,
        pt#l = Some e /\
        check_block p pt e i.

  Lemma towards_halt : forall p s i s' i',
    stepstar p s i s' i'
    -> i' = instr_halt
    -> forall s'' i'', stepstar p s i s'' i''
      -> step_or_halt p s'' i''.
    induction 1; my_simpler; [
      match goal with
        | [ H : stepstar _ _ _ _ _ |- _ ] =>
          inversion H; my_simpler
      end;
      match goal with
        | [ H : step _ _ instr_halt _ _ |- _ ] =>
          inversion H
      end
      | match goal with
          | [ H1 : step _ ?S _ _ _, H2 : stepstar _ ?S _ _ _ |- _ ] =>
            inversion H2; my_simpler
        end; [
          eauto
          | match goal with
              | [ H1 : step ?P ?S ?I _ _, H2 : step ?P ?S ?I _ _ |- _ ] =>
                generalize (step_deterministic _ _ _ _ _ _ _ H1 H2); my_simpler
            end
        ]
    ].
  Qed.

  Lemma stepN_race : forall p n s i s1 i1 s2 i2,
    stepN p n s i s1 i1
    -> stepstar_plus p s i s2 i2
      -> step_or_halt p s1 i1
      \/ stepstar p s i s2 i2
        /\ exists n', n' < n
          /\ stepN p n' s2 i2 s1 i1.
    induction 1; simpler; eauto;
      match goal with
        | [ H : stepstar_plus _ _ _ _ _ |- _ ] =>
          destruct H; inversion H; my_esimpler
      end;
      match goal with
        | [ H1 : step ?P ?S ?I _ _, H2 : step ?P ?S ?I _ _ |- _ ] =>
          generalize (step_deterministic _ _ _ _ _ _ _ H1 H2); my_simpler
      end;
      match goal with
        | [ H : stepN ?P _ ?S ?I ?S' ?I', _ : stepstar_plus ?P ?S ?I ?S' ?I' -> _ |- _ ] =>
          destruct H; [
            right; my_esimpler
            | idtac
          ]
      end;
      match goal with
        | [ H : stepstar_plus _ _ _ _ _ -> _ |- _ ] =>
          match type of H with
            | ?P -> _ =>
              assert P; [
                red; eauto
                | my_simpler; right; my_simpler; eauto with arith
              ]
          end
      end.
  Qed.

  Hint Resolve towards_halt stepN_stepstar.

  Lemma safetyN : forall p pt n n' l e s i s' i',
    n' <= n ->
    check_blocks p pt ->
    typing_dom_match p pt ->
    p#l = Some i ->
    pt#l = Some e ->
    check_env s e ->
    stepN p n' s i s' i' ->
    step_or_halt p s' i'.

    Ltac frob := my_simpler;
      repeat match goal with
               | [ H : S _ <= 0 |- _ ] => inversion H
               | [ H : stepstar_plus _ _ _ _ _ |- _ ] => destruct H; my_simpler
               | [ H : stepN _ _ _ _ _ _ |- _ ] => inversion H; [idtac]; clear H; my_simpler
             end.

    Ltac user :=
      match goal with
        [ Hbl : check_blocks _ _, Hl : _ # _ = _ |- _ ] =>
        generalize (Hbl _ _ Hl); my_simpler
      end;
      match goal with
        [ Hbl : check_block _ _ _ _, Htdm : _, Hl : _ |- _ ] =>
        generalize (Hbl Htdm _ Hl); my_simpler
      end.

    induction n; destruct n'; intros; [
      user; frob; eauto
      | frob
      | user; frob; eauto
      | user; eauto
    ];
    match goal with
      | [ H1 : stepN ?P _ ?S ?X _ _, H2 : stepstar_plus ?P ?S ?X _ _ |- _ ] =>
        generalize (stepN_race H1 H2); my_simpler
    end;
    match goal with
      | [ H : stepN _ ?N _ _ _ _ |- _ ] => eapply (IHn N); eauto with arith
    end.
  Qed.

  Lemma check_env_empty : forall e, check_env e (empty ty).
    check_env.
  Qed.

  Hint Resolve check_env_empty.

  Theorem safety : forall (p:program) (pt: program_typing),
    check_blocks p pt ->
    typing_dom_match p pt ->
    pt#entry_label = Some (empty ty) ->
    safe_prog p.
    do 2 red; unfold start_state; intros;
      match goal with
        | [ H : stepstar _ _ _ _ _ |- _ ] =>
          generalize (stepstar_stepN _ _ _ _ _ H); firstorder;
            eapply safetyN; eauto
      end.
  Qed.

End M.
